#ifndef Process_H
#define Process_H

#include "interrupt.h"

// Class: Process
// A process is the smallest executable unit in the BillOs.
// This class is somewhat analagous to a Process Control Block (PCB) in our text.
//
class Process { 

public:
	// Constructor - You can create an empty "null" process with
	// no ID or CPU requirements, or more usefully, you can create a 
	// process with an ID and CPU requirements.
	// These values are set at creation and never changed.
	// New processes are created with state="new" and with no I/O requests
	Process() ; 
	Process( int, int) ; 
	int getId() { return pId ; }	// get the process ID number

	// State methods
	// The Process::State enum defines the possible states of a process
	//   When created, the process is "new"
	//   When in the Ready Queue, it is "ready"
	//   When excuting on the CPU, it is "running"
	//   When blocked for an I/O request, it is "waiting"
	//   When its CPU time is completed, it can be "terminated"
	// Get/Set access methods are provided.
	//
	// Notes: I ended up NOT using the Process state. I left it in because
	// adding it would probably simplify my logic a bit. Later, right?
	enum State { pNew, pReady, pRunning, pWaiting, pTerminated} ;

	void	setState( State s) { pState = s ; }
	State	getState() { return pState ; }

	// Executes the Process for 1 time unit or clock tick. 
	// All Process accounting is updated to reflect the CPU time.
	// The return value is 0 if the Process executed normally.
	// The return value is an interrupt if it needs attention, for example,
	// if the Process needs I/O or has completed its CPU requirement.
	Interrupt *doCpuBurst() ; 

	// Get methods for process CPU time:
	//    CpuRequired - total time defined at process creation
	//    CpuCompleted - total time completed so far
	//    CpuRemaining - total time still neded for process to complete
	int	getCpuRequired() { return pCpuRequired ; } 
	int getCpuCompleted() { return pCpuCompleted ; }
	int	getCpuRemaining() { return pCpuRequired - pCpuCompleted ; }	

	// Method performs 1 time unit of I/O for the Process.
	// All Process accounting is kept up to date.
	// This method is similar to the doCpuBurst() method above. 
	// The return value is 0 for normal operation.
	// The return value is an interrupt if attention is required, 
	// for example, if the Process is done with I/O.
	Interrupt *doIoBurst() ; 

	// Set/Get methods for 2 possible I/O requests. An I/O request includes
	// a CPU time when it must occur, and a length. A zero value for these
	// fields means NO I/O is needed.
	// Notes: Yes, of course, I wimped out. I should have made this a List of
	// IO requests and probably added an IOEvent or IORequest class to boot.
	void setIoRequest1( int, int) ;	
	void setIoRequest2( int, int) ; 

	// Gets the CPU time/length of I/O req #1/#2. Returns -1 if there is no I/O req
	int getIoRequestTime1() { return pIoRequestTime1 ; }
	int getIoRequestLength1() { return pIoRequestLength1 ; }
	int getIoRequestTime2() { return pIoRequestTime2 ; }
	int getIoRequestLength2() { return pIoRequestLength2 ; }

	// Set/Get the termination (or completion) time for the process.
	void setTerminatedTime( int clock) { pTerminatedTime = clock ; }
	int	getTerminatedTime() { return pTerminatedTime ; }

	// Returns the total time the process waited in the ReadyQ.
	// This should only be called after Process termination time has been set.
	int calcWaitTime() ; 

private: 
	int		pId ;			// defined during creation
	State	pState ; 
	int		pCpuRequired ;	// defined during creation
	int		pCpuCompleted ;		// CPU so far
	int		pTerminatedTime ;	// CPU still needed to complete

	int		pIoRequestTime1 ;	
	int		pIoRequestLength1 ; 
	int		pIoCompleted1 ;		// I/O #1 time spent so far
	int		pIoRequestTime2 ; 
	int		pIoRequestLength2 ; 
	int		pIoCompleted2 ;		// I/O #2 time spent so far
};

#endif
