#include <iostream>
#include <fstream>
#include "billos.h"

using namespace std ; 

// Constructor
BillOs::BillOs() { 
	systemClock = 0 ; 
	timeSlice = 1 ; 
	debugLevel = 0 ; 
}

// Add the given Process to the ReadyQ
void BillOs::addProcess( Process *p) {
	readyQ.push( p) ; 
}

// Read a file of process descriptions, one per line.
// Each process is added to the ReadyQ.
// Notes: There is no error checking in this method.
void BillOs::addProcessFile( fstream &inFile, fstream &outFile) {
	char		ch ; 
	const int	maxLineLength = 100 ; 
	char		comment[ maxLineLength] ; 
	int			id, cpuRequired, ioTime1, ioLength1, ioTime2, ioLength2 ; 

	while( !inFile.eof()) { 
		if( !inFile.get( ch)) { 
			return ; 
		}
		if( ch == '#') {	// comment lines start with # sign
			inFile.getline( comment, maxLineLength) ; 
		}
		else {		// this line is not a comment, get the process info
			inFile.putback( ch) ; 

			inFile >> id ; 
			if( inFile.eof()) { return ; }
			inFile >> cpuRequired ; 
			inFile >> ioTime1 ; 
			inFile >> ioLength1 ; 
			inFile >> ioTime2 ; 
			inFile >> ioLength2 ;

			outFile << id << "\t" << cpuRequired << "\t" ;
			outFile << ioTime1 << "\t" << ioLength1 << "\t" ; 
			outFile << ioTime2 << "\t" << ioLength2 << endl ; 

		// Create a new process with this information
			Process *p = new Process( id, cpuRequired) ; 
			p->setIoRequest1( ioTime1, ioLength1) ; 
			p->setIoRequest2( ioTime2, ioLength2) ; 

			readyQ.push( p) ; 
		}
	}
}

// Initialize the OS prior to running the simulation.
// Pop the first process off the Ready Q and place it on the CPU.
// Set the system clock to 0... and we're ready to go.
void BillOs::init() { 
	Process *p ; 

	p = readyQ.front() ; 
	readyQ.pop() ; 
	theCpu.setRunningProcess( p) ; 

	systemClock = 0 ; 
}

// Method runs the simulation to its gnarly completion
// After this method, I suggest you call outputStats()
void BillOs::run() {
	while( ! doneRunning()) { 
		runForOneTick() ; 
	}
}

// Notes: Did not complete this, but it's easy to do!
void BillOs::run( int) {
	cout << "Coming soon to a theatre near you!" ; 
}

// Notes: Did not complete this, but its easy to do!
void runTillNextInterrupt() {
	cout << "Coming soon to a theatre near you!" ; 
}

// Returns true if the OS is completed all its tasks and is done running
bool BillOs::doneRunning() { 
	if( !theCpu.isBusy()  &&
		readyQ.empty()  && ioList.empty()) { 
		return true ; 
	}
	return false ; 
}

// Output the stats report for Program #1. Basically, just the termination
// time and wait for each process.
void BillOs::outputStats( fstream &outFile) { 
	list<Process *>::iterator	iter ;
	Process						*p ; 

	// Print out the stats of all the completed/terminated processes
	outFile << "\n\nSTATS:" << endl ; 
	iter = terminatedList.begin() ; 
	while( iter != terminatedList.end()) { 
		p = *iter ; 

		outFile << "Process " << p->getId() ; 
		outFile << " completed at time " << p->getTerminatedTime() ; 
		outFile << " with a wait of " << p->calcWaitTime() << endl ; 

		iter++ ; 
	}
}

//////////////////////////////////////////////////////////
// Private methods
//////////////////////////////////////////////////////////

// Simulate the BillOs for one tick
// 
void BillOs::runForOneTick() { 
	systemClock++ ; 

	// Notes: The order here is important, and this is a cleeb.
	// Running the CPU may place a process on the I/O List, so we have to 
	// do I/O first. The correct implementation wouuld be to have both the
	// CPU and I/O devices create a list of Interrupts that would be handled
	// after the work is done for this "tick". Then Processes would be moved
	// around and we wouldn't have to worry about it.
	runIO() ; 
	runCPU() ; 
}

// This method does one tick-worth of I/O
// Notes: I/O devices should be handled the same way that CPU's are.
// There should be a I/O device class that is called that generates interrupts.
void BillOs::runIO() { 
	Process		*p ; 
	Process		*removeP ; 
	Interrupt	*intr ; 
	list<Process *>::iterator	iter ; 

	if( ioList.empty()) { return ; }

	iter = ioList.begin() ; 
	while( iter != ioList.end()) { 
		removeP = 0 ; 
		p = *iter ; 
		intr = p->doIoBurst() ; 

		if( intr != 0) {
//			debugInterrupt( int, p, systemClock, outFile) ; 

			switch( intr->getType()) { 
			case Interrupt::IoComplete:
				// Notes: Interesting note here... the STL barfed when 
				// I tried to erase the current item in the list. This is
				// in direct opposition to the documentation I have, but so be it.
				// I just save the item and erase it after incrementing the iterator,
				// and that worked fine.
				removeP = p ; 
//				ioList.erase( ioIter) ; 
				readyQ.push( p) ; 
				break ; 
			default: 
				cout << "Error: Bad I/O interrupt, type=" << intr->getType() << endl ; 
				break ; 
			}
		}
		if( iter != ioList.end()) { 
			iter++ ; 
		}
		if( removeP != 0) { 
			ioList.remove( removeP) ; 
			removeP = 0 ; 
		}
	}
}

// This method does one tick-worth of CPU.
// In an Interrupt is generated during this tick, then it is handled here.
void BillOs::runCPU() { 
	Interrupt		*intr ; 
	Process			*p ; 

	// Run the CPU process for one time unit
	intr = theCpu.doBurst() ; 
	if( intr == 0) { 
		// Check if the time slice is up for this process
		if( theCpu.getRunTotal() == timeSlice) { 
			intr = new Interrupt( Interrupt::TimeSliceComplete) ; 
		}
	}

	// Switch handles all interrupts that can be generated when 
	// a process is running on the CPU
	if( intr != 0) {
		p = theCpu.getRunningProcess() ; 
		theCpu.removeRunningProcess() ; 

//		debugInterrupt( intr, p, systemClock, outFile) ; 

		switch( intr->getType()) { 
		case Interrupt::CpuComplete:
			p->setTerminatedTime( systemClock) ; 
			terminatedList.push_back( p) ; 
			break ; 

		case Interrupt::IoRequest:
			ioList.push_back( p) ; 
			break ; 

		case Interrupt::TimeSliceComplete:
			readyQ.push( p) ; 
			break ; 
				
		default: 
			cout << "Error: Bad CPU interrupt, type=" << intr->getType() << endl ; 
			break ; 
		}
	}

	// If, after all is said and done, the CPU is not busy, then 
	// pop a process off the ReadyQ
	if( !theCpu.isBusy()) { 
		if( !readyQ.empty()) { 
			p = readyQ.front() ; 
			readyQ.pop() ; 
			theCpu.setRunningProcess( p) ; 
		}
	}
}

#ifdef ZZZ_LATER
// Debug an interrupt
void BillOs::debugInterrupt( Interrupt *intr, Process *p, int clock, 
						   fstream &outFile) {
	if( debugInts) { 
		outFile << systemClock << ": " ; 
		outFile << "p(" << p->getId() << ") " ; 
		outFile << intr->getTypeString() << " intr" ; 
		outFile << " - cpu(" << p->getId() << ")= " << p->getCpuCompleted() << endl ; 
	}
}
#endif
