#ifndef BILLOS_H
#define BILLOS_H

#include <queue>
#include <list>

using namespace std ; 

#include "process.h"
#include "cpu.h"
//typedef queue<Process *> ProcessQ ; 

// Class: BillOs
// BillOs simulates an operating system... scheduling Processes for CPU 
// and I/O according to their definintion.
// Elements of the BillOs are:
//    ReadyQ - the queue of Processes ready for CPU time
//    CPU - a CPU object to execute one Process at a time
//    IOList - a list of Processes currently doing I/O
//
// BillOs operates 1 clock tick at a time. For each tick, the Cpu and all
// I/O devices are "executed" for 1 time unit. Processes communicate their
// needs to BillOs via Interrupt objects that are returned from the CPU or
// during I/O. BillOs munges these interrupts, taking the appropriate
// action for the hailing Process.
//
// Notes: There are many shortcuts here. Here are a few:
// 1) The CPU and I/O should be done through the same mechanism and probably
// even class. An IoDevice class should probably have been created that would
// have been similar to the Cpu class.
// 2) Interrupts are just an enum not. They should have included the Process 
// the raised the Interrupt, but I did not want a circular class dependency.
// If Cpu and IoDevice classes were properly defined, they would "know" about
// Interrupts, but the Process class would not.
class BillOs { 
public:
	// Constructor - no options... a blank OS awaiting your command
	BillOs() ; 

	// Set/Get the OS time slice. This determine the amount of CPU time a 
	// process receives before being pre-empted for the next process
	// waiting on the ReadyQ
	void	setTimeSlice( int ts) { 
		if( ts < 1) { 
			ts = 1 ;	// don't allow a time slice < 1 time unit!
		}
		timeSlice = ts ; 
		}
	int		getTimeSlice() { return timeSlice ; }

	// Set/Get the OS debugging level to turn debugging on or off.
	// A zero level of debugging (the default) means no debugging messages
	// The higher the debugging level, the more debugging output.
	// Notes: I didn't end up using this.
	void	setDebugLevel( int dl) { debugLevel = dl ; }
	int		getDebugLevel() { return debugLevel ; }

	// Add process(e) to the OS. New processes are put on the ReadyQ
	// Process may be added individually or via a datas file
	// Currently, you should add all the processes before running the OS
	// as that's all I have tested.
	// Notes: addProcessFile() probably doesn't belong. Why is the OS reading 
	// stupid data files?
	void	addProcess( Process *) ; 
	void	addProcessFile( fstream &, fstream &) ; 

	// You must call init() prior to any of the simulation methods below.
	// Notes: This should be removed. Check for initialization in the run()
	// methods automatically. 
	void	init() ; 

	// Simulation methods to run the OS for a number of time units
	// You can run() until the simulation is complete. 
	// You can run( int) for a specified numbner of time units.
	// Or, you can runTillNextInterrupt() and watch the simulation 
	// blow by blow.
	void	run() ; 
	void	run( int) ; 
	void	runTillNextInterrupt() ; 

	// Returns true when the BillOs simulation is done, or more precisely
	// has no more "work" to do.
	bool	doneRunning() ; 

	int		getSystemClock() { return systemClock ; }

	// Outputs the stats for Program #1 compliance.
	// Notes: Again, this shouldn't be in this class. BillOs should 
	// provide a List of Processes that it munged and reports should be 
	// generated from that. This is a shortcut.
	void	outputStats( fstream &outFile) ; 

private:
	int		timeSlice ; 
	int		debugLevel ; 
	int		systemClock ; 
	
	Cpu					theCpu ; 
	queue<Process *>	readyQ ; 
	list<Process *>		ioList ; 
	list<Process *>		terminatedList ; 

	void	runForOneTick() ; 
	void	runIO() ; 
	void	runCPU() ; 

} ; 

#endif
