#include <iostream.h>
#include <fstream>
#include <math.h>

using namespace std ; 

#include "vmsimulator.h"

void	doFifoPageReplacement( int) ; 
void	doLruPageReplacement( int) ; 


//
//
VMSimulator::VMSimulator( int addressBits, int numFrames, 
						 PageReplacementAlgorithm algorithm) { 
	vAddressBits = addressBits ;
	vNumFrames = numFrames ; 
	vAlgorithm = algorithm ; 

	vMemorySize = (int) pow( 2.0, (double) addressBits) ; 

	cout << "vms: bits=" << vAddressBits << ", frames=" << vNumFrames ; 
	cout << ", algo=" << vAlgorithm << ", vm size=" << vMemorySize << endl ; 

	// Allocate the page table. The new page table will be empty.
	vPageTable = new PageTable( vAddressBits) ; 
	cout << "vms: pagetable ok" << endl ; 

	// Allocate the "free" frames list. The frames will be numbered 1 - n.
	vFrameList = new FrameList( vNumFrames) ; 
	cout << "vms: framelist ok" << endl ; 

	vNumPageFaults = 0 ; 
}

///////
//
bool	VMSimulator::simulatePageReferences( vector<int> pageRefs,
											fstream &outFile) {
	int		page ; 
	int		frame ; 
	int		victim ; 

	for( int i = 0; i < pageRefs.size(); i++) { 

		page = pageRefs[i] ;	// current page reference

		// Before anything, make sure the page reference is in range
		// If it's bad, then report it and then pull the plug.
		if( !vPageTable->isPageInRange( page)) { 
			outFile << "Error: Bad page reference=" << page << endl ; 
			return false ; 
		}

		// Case #1. If this page already has a valid frame, in memory, 
		// then we're done.
		frame = vPageTable->getFrame( page) ; 

		if( frame < 0) { 
			// Case #2. So, the page isn't in memory. If there's a free frame,
			// then grab it and we're done.
			frame = vFrameList->allocFrame() ; 
			if( frame >= 0) { 
				vPageTable->setFrame( page, frame) ; 
//				cout << "sim: Case #2 page=" << page << " alloced frame=" << frame << endl ; 
			}
			else { 
				// Step #3. Page fault case! 
				// We need to replace a VICTIM page that's in memory and grab his frame
				switch( vAlgorithm) { 
				case FIFO: 
					victim = findFifoVictim( page) ; 
					break ; 

				case LRU: 
					victim = findLruVictim( page) ;
					break ; 

				case Optimal: 
					victim = findOptimalVictim( page, pageRefs, i) ; 
					break ; 

				default: 
					cout << "vms: Way bad enum in simulation" << endl ; 
					break ; 
				}

				// Now that we have the victim, take his frame and give it
				// to the new page reference.
				frame = vPageTable->getFrame( victim) ; 
				vPageTable->unsetFrame( victim) ; 
				vPageTable->setFrame( page, frame) ; 
			}

			vFifoQ.push( page) ; 
			vNumPageFaults++ ; 
		}

		vPageTable->setTimeOfUse( page, i) ; 

		outFile << "Page table (page,frame): " ; 
		vPageTable->debug( outFile) ; 
		outFile << endl ; 
//		cin.get( ch) ;			// pause at bottom of the loop
	}

	return true ;		// we successfully processed all the page refs
}

//
//
void VMSimulator::report( fstream &outFile) { 
	int			frame ; 

	outFile << "   Page Faults: " << vNumPageFaults << endl ; 
	outFile << "   Page Table:" << endl ; 

	for( int p = 0; p < vMemorySize; p++) { 
		frame = vPageTable->getFrame( p) ; 
		if( frame >= 0) { 
			outFile << "      Page: " << p << "   Frame: " << frame << endl ; 
		}
	}
}

///////
// 
int VMSimulator::findFifoVictim( int page) { 
	int		victim = vFifoQ.front() ; 
	vFifoQ.pop() ; 
	return victim ; 
}

///////
// findLruVictim
//
int VMSimulator::findLruVictim( int page) { 
	int		curTime ; 
	int		lruVictim = -1 ; 
	int		lruTime = vPageTable->getSize() ; 

	// Find the victim. He's the page with the smallest timeOfUse
	// field in the Page Table.
	for( int curPage = 0; curPage < vPageTable->getSize() ; curPage++) { 
		curTime = vPageTable->getTimeOfUse( curPage) ; 
		if( curTime != -1  &&  curTime < lruTime) {
			lruVictim = curPage ; 
			lruTime = curTime ; 
		}
	}

	return lruVictim ; 
}

///////
// findOptimalVictim
//
int VMSimulator::findOptimalVictim( int page, 
								   vector<int> pageRefs, int curRef) {
	int		optVictim = -1 ; 
	int		optTime ; 
	int		nextRefTime ; 

	// Loop through all the valid pages (those with frames) in the
	// page table. Select the optimal victim... the one used furthest
	// in the future (per our page reference string)
	for( int curPage = 0; curPage < vPageTable->getSize() ; curPage++) { 

		if( vPageTable->isPageValid( curPage)) { 
			// Find the next time in the future that this page is ref'd
			nextRefTime = -1 ; 
			for( int i = curRef+1; i < pageRefs.size(); i++) { 
				if( pageRefs[i] == curPage) { 
					nextRefTime = i ; 
					break ; 
				}
			}

			if( nextRefTime == -1) { 
				// This page is never used again, so it's the victim
				// and we're done
				optVictim = curPage ; 
				break ; 
			}
			else if( nextRefTime > optTime) { 
				// This page is used further in the future than our
				// current victim choice, so make it the prospective
				// victim. We're not done yet though, keep lokoing 
				// for a better (optimal) victim
				optVictim = curPage ; 
				optTime = nextRefTime ; 
			}
		}
	}

	return optVictim ; 	
}
